@@ -6,10 +6,12 @@ Licensed under MIT
 
 VERSION HISTORY
 - v1.3: order book helper functions (p_marg_f, yfromp_f, dyfromp_f, dyfromdx_f, dyfromdx_f, goalseek)
-- v1.3.1: more order book and other helper functions (yfromx_f, xfromy_f, p_eff_f, xint, )
+- v1.4: new methods: fromQxy, Q, Gamma, sellx, selly; yfromx_f, xfromy_f, p_eff_f, xint (1.3.1)
+- v1.5: linked curves (beta); bidask, curves_by_pair_bidask, checks for B,S=0 (1.4.1)
+- v1.6: linked curves incl trading (final), addliqy, tradeto; minor formula improvement (1.5.1)
 """
-__version__ = "1.3.1"
-__date__ = "13/Dec/2022"
+__version__ = "1.6"
+__date__ = "20/Jan/2023"
 
 try:
     from .pair import CarbonPair
@@ -31,7 +33,13 @@ class CarbonOrderUI:
     :S:       the S-parameter; S = sqrt pa_raw - Sqrt pb_raw
     :yint:    the y-intercept of the curve (also its current maximum capacity)
     :y:       the current y-coordinate on the curve (also current token holdings)
-    other properties
+    :id:      id of this curve*
+    :linked:  the linked curve object*
+
+    *Beta; interace may still change in respect to ix, lix and linked as well 
+    as the associated functions
+    
+    other properties set by the constructor
     :pa_raw:   the pa parameter in native quotation (dy/dx)
     :pb_raw:   ditto pb
     :pa:       the pa paramter in the quotation appropriate for the pair
@@ -57,6 +65,9 @@ class CarbonOrderUI:
     S: float
     yint: float
     y: float
+    id: any = None
+    linked: any = None
+
     def __post_init__(self):
         self.pair = CarbonPair(self.pair)
         self.tkn = self.tkn.upper()
@@ -85,6 +96,29 @@ class CarbonOrderUI:
             self.pmin = None
             self.pmax = None        
     
+    def set_id(self, id):
+        """sets curve index; raises if curve ID already set"""
+        if not self.id is None:
+            raise ValueError("Curve ID has already been set", id, self.id)
+        self.id = id
+    
+    @property
+    def lid(self):
+        """returns ID of linked curve, or None"""
+        if self.linked is None:
+            return None
+        return self.linked.id
+        
+    def set_linked(self, linked=None):
+        """
+        sets linked object and index
+
+        :linked:    the linked curve (CarbonOrderUI object)
+        """
+        if not self.linked is None:
+            raise ValueError("Linked object has already been set", linked.id, self.linked.id)
+        self.linked = linked
+
     @classmethod
     def from_BSy(cls, pair, tkn, B, S, yint, y):
         """
@@ -112,7 +146,7 @@ class CarbonOrderUI:
     @classmethod
     def from_prices(cls, pair, tkn, pa, pb, yint, y):
         """
-        alternative constructor, taking prices pa, pb
+        alternative constructor, taking prices pa, pb and curve capacity yint
         
         :pair:    the corresponding token pair (specifically, its CarbonPair record)
         :tkn:     the token that this order is selling
@@ -149,27 +183,81 @@ class CarbonOrderUI:
             pair=pair, 
             tkn=tkn, 
             B=B, 
-            S= S, 
+            S=S, 
             yint=yint, 
             y=y
         )
     
+    @classmethod
+    def from_Qxy(cls, pair, tkn, Q, xint, yint, y):
+        """
+        alternative constructor, taking convexity Q and curve capacities xint, yint
+        """
+        pair = CarbonPair(pair)
+        if not pair.has_token(tkn):
+            raise ValueError("token not part of the pair", tkn, pair)
+        if xint<=0:
+            raise ValueError("xint must be positive", xint)
+        if yint<=0:
+            raise ValueError("yint must be positive", yint)
+        if y>yint:
+            raise ValueError("y must not be bigger than yint (y={y}, yint={yint})", yint, y)
+        if y<0:
+            raise ValueError("y must be non-negative", y)
+
+        p0 = yint/xint
+        B = sqrt(p0*Q)
+        S = sqrt(p0/Q) - B
+
+        return cls(
+            pair=pair, 
+            tkn=tkn, 
+            B=B, 
+            S=S, 
+            yint=yint, 
+            y=y
+        )
+
     @classmethod
     def from_order(cls, order):
         """
         alternative constructor, from an Order object
 
-        :order:     the order object
+        :order:     the Order or CarbonOrderUI object
         """
+        try: 
+            yint = order.y_int
+        except:
+            yint = order.yint
+
         return cls(
             pair=order.pair, 
             tkn=order.tkn, 
             B=float(order.B), 
             S= float(order.S), 
-            yint=float(order.y_int), 
-            y=float(order.y)
+            yint=float(yint), 
+            y=float(order.y),
+            id=order.id,
         )
 
+    @property 
+    def tkny(self):
+        """the token on the y-axis (alias for tkn)"""
+        return self.tkn
+    
+    @property 
+    def tknx(self):
+        """the token on the x-axis"""
+        return self.pair.other(self.tkn)
+
+    @property
+    def bidask(self):
+        """returns BID or ASK depending on what type curve it is"""
+        if self.tkny == self.pair.basetoken:
+            return "BID"
+        else:
+            return "ASK"
+    
     @property
     def p0(self):
         """
@@ -181,12 +269,24 @@ class CarbonOrderUI:
     def px(self):
         """alias for pb"""
         return self.pb 
-    
+    p_end = px
+
     @property
     def py(self):
         """alias for pa"""
-        return self.pa  
+        return self.pa
+    p_start = py
     
+    @property
+    def Q(self):
+        """Q parameter = sqrt(Px/Py) < 1"""
+        return sqrt(self.pb_raw/self.pa_raw)
+
+    @property
+    def Gamma(self):
+        """Gamma parameter (also known as n) = 1 - sqrt(Q)"""
+        return 1 - sqrt(self.Q)
+
     @property
     def widthr(self):
         """the width ratio of the range, widthr = pmax/pmin"""
@@ -303,6 +403,11 @@ class CarbonOrderUI:
         result = dydx if self.pair.has_quotetoken(self.tkn) else 1/dydx
         return result
     
+    class PriceOutOfBoundsError(ValueError): pass
+    class PriceOutOfBoundsErrorBeyondStart(PriceOutOfBoundsError): pass
+    class PriceOutOfBoundsErrorBeyondEnd(PriceOutOfBoundsError): pass
+    class PriceOutOfBoundsErrorBeyondMarg(PriceOutOfBoundsError): pass
+
     def yfromp_f(self, p, checkbounds=True, raiseonerror=False):
         """
         returns y as a function of the target marginal price
@@ -321,26 +426,27 @@ class CarbonOrderUI:
         if checkbounds:
             if dydx > self.pa_raw:
                 if raiseonerror:
-                    raise ValueError("Price out of bounds (beyond start)", p, self.pa)
+                    raise self.PriceOutOfBoundsErrorBeyondStart("Price out of bounds (beyond start)", p, self.pa)
                 return None
             elif dydx < self.pb_raw:
                 if raiseonerror:
-                    raise ValueError("Price out of bounds (beyond end)", p, self.pb)
+                    raise self.PriceOutOfBoundsErrorBeyondEnd("Price out of bounds (beyond end)", p, self.pb)
                 return 0
         y = self.yint * (sqrt(dydx) - self.B) / self.S
         if checkbounds:
             if y > self.y:
                 if raiseonerror:
-                    raise ValueError("Price out of bounds (beyond marginal), hence target y > y", y, self.y )
+                    raise self.PriceOutOfBoundsErrorBeyondMarg("Price out of bounds (beyond marginal), hence target y > y", y, self.y )
                 return None
         return y
 
     @property
     def xint(self):
         """
-        x intercept, ie max x liquidit, ie x at y=0
+        x intercept, ie max x liquidity, ie x at y=0
         """
-        return self.yint**2 / (self.B**2*self.yint + self.B*self.S*self.yint)
+        #return self.yint**2 / (self.B**2*self.yint + self.B*self.S*self.yint)
+        return self.yint / (self.B**2 + self.B*self.S)
         
     def dyfromp_f(self, p, checkbounds=True, raiseonerror=False):
         """
@@ -352,6 +458,10 @@ class CarbonOrderUI:
         :returns:       the (positive!) dy value at which the marginal price is achieved
                         in cases where yfromp_f returns none, this func returns 0
         """
+        if self.B == 0 and self.S == 0:
+            if raiseonerror:
+                raise ValueError("Can't determine trade prices from an empty curve", self.B, self.S)
+            return 0
         y = self.yfromp_f(p, checkbounds, raiseonerror)
         if y is None: return 0
         return self.y-y
@@ -369,6 +479,11 @@ class CarbonOrderUI:
         AMM SELLing x. In this case it returns a negative number, corresponding
         to the AMM BUYing y.
         """
+        if self.B == 0 and self.S == 0:
+            if raiseonerror:
+                raise ValueError("Can't trade on an empty curve", self.B, self.S)
+            return 0
+
         if checkbounds:
             if dx < 0:
                 if raiseonerror:
@@ -405,6 +520,11 @@ class CarbonOrderUI:
         AMM BUYing y. In this case it returns a negative number, corresponding
         to the AMM SELLing x.
         """
+        if self.B == 0 and self.S == 0:
+            if raiseonerror:
+                raise ValueError("Can't trade on an empty curve", self.B, self.S)
+            return 0
+
         if checkbounds:
             if dy < 0:
                 if raiseonerror:
@@ -623,3 +743,180 @@ class CarbonOrderUI:
             #print(f"m={m}, m={m}, b={b}")
             if b/a-1 < eps:
                 return m
+
+    def selly(self, dy, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
+        """
+        executes a trade selling dy for dx (dy given)
+
+        :dy:            the amount of y to sell (a positive number)
+        :execute:       if False, only display results but do not update the object
+        :allowneg:      if True, negative dy numbers (=buying) are allowed
+        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
+                        only meaningful with allowneg = True
+        :raiseonerror:  if True, error lead to raising on exception
+        """
+        if dy < 0:
+            if not allowneg:
+                if raiseonerror:
+                    raise ValueError(f"Negative dy is not allowed (allowneg={allowneg})", dy)
+                return None
+
+        if dy == 0:
+            execute = False
+
+        yold = self.y
+        pold = self.p_marg
+        ynew = self.y - dy
+        pnew = self.p_marg_f(dy, checkbounds=False)
+
+        if ynew < 0:
+            if raiseonerror:
+                raise ValueError(f"Traded beyond capacity (yold={yold}, ynew={ynew}, dy={dy})")
+            return None
+
+        elif ynew > self.yint:
+            if not expandcurve:
+                if raiseonerror:
+                    raise ValueError(f"Curve needs expanding and expanding not allowed (yold={yold}, ynew={ynew}, dy={dy}, expandcurve={expandcurve})")
+                return None
+            else:
+                yintold = self.yint
+                yint = ynew
+                if execute:
+                    self.yint = ynew
+                curve_expanded = True
+        
+        else:
+            curve_expanded = False
+            yint = self.yint
+
+        dx = self.dxfromdy_f(dy, checkbounds=False, raiseonerror=True)
+        if execute:
+            self.y = ynew
+
+        if self.linked:
+            result_linked = self.linked.addliqy(dx)
+        else:
+            result_linked = None
+
+        result = {
+            "action":       "bysource[selly]",
+            "executed":     execute,
+            "y_old":        yold,
+            "y":            ynew,
+            "dy":           dy,
+            "yint_old":     yintold if curve_expanded else None,
+            "y_int":        yint,
+            "expanded":     curve_expanded,
+            "x":            self.xfromy_f(ynew),
+            "dx":           dx,
+            "tkny":         self.tkny,
+            "tknx":         self.tknx,
+            "tx":           f"Sell {abs(dy)} {self.tkny} buy {self.tknx}" if dx>0 else f"Buy {abs(dy)} {self.tkny} sell {self.tknx}",
+            "dx/dy":        dx/dy if dy != 0 else None,
+            "dy/dx":        dy/dx if dx != 0 else None,
+            "pmarg_old":    pold,
+            "pmarg":        pnew,
+            "linked":       result_linked,
+            "p":            None,
+        } 
+        result["p"] = result["dx/dy"] if self.pair.has_basetoken(self.tkny) else result["dy/dx"]
+        return result
+
+    def buyx(self, dx, execute=True, allowneg=True, expandcurve=False, raiseonerror=False):
+        """
+        executes a trade dy for dx (dx given)
+
+        :dx:            the amount of x to buy (a positive number)
+        :execute:       if False, only display results but do not update the object
+        :allowneg:      if True, negative dy numbers (=buying) are allowed
+        :expandcurve:   if True, purchasing y beyond yint [sic] expands the curve to yint = y
+                        only meaningful with allowneg = True
+        :raiseonerror:  if True, error lead to raising on exception
+        """
+        dy = self.dyfromdx_f(dx, checkbounds=False, raiseonerror=True)
+        result = self.selly(dy, execute, allowneg, expandcurve, raiseonerror)
+        result["action"] = "bytarget[buyx]"
+        return result
+
+    def tradeto(self, p_marg, execute=True, raiseonerror=False):
+        """
+        executes a trade dy for dx (target marginal price p_marg given)
+
+        :p:             the target marginal price
+        :execute:       if False, only display results, but do not update the object
+        :raiseonerror:  if True, error lead to raising on exception
+        """
+        try:
+            dy = self.dyfromp_f(p_marg, checkbounds=True, raiseonerror=True)
+        except (self.PriceOutOfBoundsErrorBeyondStart, 
+                self.PriceOutOfBoundsErrorBeyondMarg):
+            dy = 0
+        except self.PriceOutOfBoundsErrorBeyondEnd:
+            dy = self.dyfromp_f(self.p_end, checkbounds=False, raiseonerror=False)
+        result = self.selly(dy, execute, allowneg=False, expandcurve=False, raiseonerror=raiseonerror)
+        result["action"] = "byprice[tradeto]"
+        return result
+        
+    def addliqy(self, dy, expandcurve=True):
+        """
+        adds liquidity to the curve (typically called on a linked curve)
+
+        :dy:            the amount of liquidity to be added (must be positive)
+        :expandcurve:   if True (default), expand yint=y if need be
+        """
+        if dy < 0:
+            raise ValueError("Liquidity amount dy must not be < 0", dy)
+        
+        newy = self.y + dy
+        result = {
+            "y_old":        self.y,
+            "y":            newy,
+            "dy":           dy,
+            "yint_old":     self.yint,
+            "yint":         self.yint,
+            "tkny":         self.tkn,
+            "pmarg_old":    self.p_marg,
+            "pmarg":        None,
+            "expanded":     False,
+        }
+        if newy > self.yint:
+            if not expandcurve:
+                raise ValueError("Must expand curve and expandcurve=False", self.y, dy, newy, self.yint)
+            self.yint = newy
+            result["yint"] = newy
+            result["expanded"] = True
+        self.y = newy
+        result["pmarg"] = self.p_marg
+        return result
+        
+    @staticmethod
+    def curves_by_pair_bidask(curves):
+        """
+        sorts curves by pair, and then by bid/ask
+
+        :curves:        an iterable of CarbonOrderUI curves
+        :returns:       dict {
+                            "pair": {
+                                "ALL": [curve, ...],
+                                "BID": [curve, ...],
+                                "ASK": [curve, ...]
+                            }
+                            ...
+                        }
+        """
+        pairs = set(r.pair.slashpair for r in curves.values())
+        result = {
+            pair:{ 
+                "ALL": [v for v in curves.values() if v.pair.slashpair == pair],
+                "ASK": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="BID"],
+                "BID": [v for v in curves.values() if v.pair.slashpair == pair and v.bidask=="ASK"],
+            }
+            for pair in pairs
+        }            
+        return result
+    
+    def __repr__(self):
+        s1 = f"pair={self.pair.slashpair}, tkn={self.tkn}, B={self.B}, S={self.S}, yint={self.yint}, y={self.y}, id={self.id}"
+        s2 = f"linked=<{self.linked.id}>" if self.linked else "linked=None"
+        return f"{self.__class__.__name__}({s1}, {s2})" 
\ No newline at end of file
